
#include <stdio.h>
#include <stdlib.h>
#include "gdal_priv.h"
#include "gdal_rat.h"

int anIntData[] = {197, 83, 46, 29, 1, 78, 23, 90, 12, 45 };
double adfDoubleData[] = {0.1, 43.2, 78.1, 9.9, 23.0, 0.92, 82.5, 0.0, 1.0, 99.0};
char *apszStringData[] = {(char*)"sddf", (char*)"wess", (char*)"grbgr", (char*)"dewd", (char*)"ddww", (char*)"qwsqw", (char*)"gbfgbf", (char*)"wwqw3", (char*)"e", (char*)""};
char *apszStringDataInts[] = {(char*)"197", (char*)"83", (char*)"46", (char*)"29", (char*)"1", (char*)"78", (char*)"23", (char*)"90", (char*)"12", (char*)"45"};
char *apszStringDataDoubles[] = {(char*)"0.1", (char*)"43.2", (char*)"78.1", (char*)"9.9", (char*)"23.0", (char*)"0.92", (char*)"82.5", (char*)"0.0", (char*)"1.0", (char*)"99.0"};

char *apszLongStringData[] = {(char*)"sdfsdfsdfs", (char*)"sdweddw", (char*)"sdewdweee", (char*)"3423dedd", 
                            (char*)"jkejjjdjd", (char*)"edcdcdcdc", (char*)"fcdkmk4m534m", (char*)"edwededdd",
                            (char*)"dedwedew", (char*)"wdedefrfrfrf"};

void errexit(const char *pszString)
{
    fprintf( stderr, "%s\n", pszString);
    exit(1);
}

void CreateAndWriteRAT(const char *pszName)
{
    GDALDriver *poDriver;
    // Create our test dataset
    poDriver = GetGDALDriverManager()->GetDriverByName("HFA");

    GDALDataset *poDS;
    poDS = poDriver->Create(pszName, 10, 10, 1, GDT_Byte, NULL );

    if( poDS->GetAccess() == GA_ReadOnly )
        errexit("file not writeable");

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // write some image data
    GByte *pData = (GByte*)CPLCalloc(sizeof(GByte), 10 * 10);
    pData[20] = 20; // so we can see the whole RAT in Imagine
    poBand->RasterIO(GF_Write, 0, 0, 10, 10, pData, 10, 10, GDT_Byte, 0, 0);
    CPLFree(pData);

    // set the transform
    double afdTransfrom[6] = {0, 1, 0, 0, 0, -1};
    poDS->SetGeoTransform(afdTransfrom);
    poDS->SetMetadataItem("LAYER_TYPE", "thematic");

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    int nRows = sizeof(anIntData) / sizeof(int);
    //poRAT->SetRowCount(nRows);

    // create columns
    if( poRAT->CreateColumn("Ints", GFT_Integer, GFU_Generic ) != CE_None )
        errexit("Failed to create column");
    if( poRAT->CreateColumn("Doubles", GFT_Real, GFU_Generic ) != CE_None )
        errexit("Failed to create column");
    if( poRAT->CreateColumn("Strings", GFT_String, GFU_Generic ) != CE_None )
        errexit("Failed to create column");

    // for writing as different type
    if( poRAT->CreateColumn("IntsAsDouble", GFT_Integer, GFU_Generic ) != CE_None )
        errexit("Failed to create column");
    if( poRAT->CreateColumn("IntsAsString", GFT_Integer, GFU_Generic ) != CE_None )
        errexit("Failed to create column");

    if( poRAT->CreateColumn("DoubleAsInt", GFT_Real, GFU_Generic ) != CE_None )
        errexit("Failed to create column");
    if( poRAT->CreateColumn("DoubleAsString", GFT_Real, GFU_Generic ) != CE_None )
        errexit("Failed to create column");


    poRAT->SetRowCount(nRows);

    if( poRAT->GetRowCount() != nRows)
        errexit("Wrong RowCount");
    if( poRAT->GetColumnCount() != 7 )
        errexit("Wrong columncount");
    if( !EQUAL(poRAT->GetNameOfCol(1), "Doubles"))
        errexit("Wrong column name");
    if( poRAT->GetUsageOfCol(1) != GFU_Generic)
        errexit("Wrong column usage");
    if( poRAT->GetTypeOfCol(1) != GFT_Real )
        errexit("Wrong column type");
    if( poRAT->GetColOfUsage(GFU_Generic) != 0)
        errexit("Wrong col of usage");
    if( !poRAT->ChangesAreWrittenToFile())
        errexit("Wrong ChangesAreWrittenToFile");

    // Write data
    if( poRAT->ValuesIO(GF_Write, 0, 0, nRows, anIntData) != CE_None )
        errexit("Failed to write int column");
    if( poRAT->ValuesIO(GF_Write, 1, 0, nRows, adfDoubleData) != CE_None )
        errexit("Failed to write double column");
    if( poRAT->ValuesIO(GF_Write, 2, 0, nRows, apszStringData) != CE_None )
        errexit("Failed to write string column");

    // different types
    if( poRAT->ValuesIO(GF_Write, 3, 0, nRows, adfDoubleData) != CE_None )
        errexit("Failed to write doubles to int column");
    if( poRAT->ValuesIO(GF_Write, 4, 0, nRows, apszStringDataInts) != CE_None )
        errexit("Failed to write strings to int column");

    if( poRAT->ValuesIO(GF_Write, 5, 0, nRows, anIntData) != CE_None )
        errexit("Failed to write ints to doubles column");
    if( poRAT->ValuesIO(GF_Write, 6, 0, nRows, apszStringDataDoubles) != CE_None )
        errexit("Failed to write strings to doubles column");


    fprintf( stderr, "Succeeding writing data\n" );

    GDALClose( (GDALDatasetH) poDS );
}

int CheckArraysInt(int *p1, int *p2, int nLength)
{
    for( int i = 0; i < nLength; i++ )
    {
        if( p1[i] != p2[i] )
        {
            fprintf( stderr, "%d != %d %d\n", p1[i], p2[i], i);
            return FALSE;
        }
    }
    return TRUE;
}

int CheckArraysDouble(double *p1, double *p2, int nLength)
{
    for( int i = 0; i < nLength; i++ )
    {
        if( p1[i] != p2[i] )
            return FALSE;
    }
    return TRUE;
}

int CheckArraysDoubleAsInt(double *p1, double *p2, int nLength)
{
    for( int i = 0; i < nLength; i++ )
    {
        if( int(p1[i]) != int(p2[i]) )
            return FALSE;
    }
    return TRUE;
}

int CheckArraysString(char **p1, char **p2, int nLength)
{
    for( int i = 0; i < nLength; i++ )
    {
        if( !EQUAL(p1[i], p2[i]) )
            return FALSE;
    }
    return TRUE;
}

int CheckArraysStringAsDouble(char **p1, char **p2, int nLength)
{
    for( int i = 0; i < nLength; i++ )
    {
        if( atof(p1[i]) !=  atof(p2[i]) )
            return FALSE;
    }
    return TRUE;
}

void ReadAndCheckValues(const char *pszName, int nNumRows)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_ReadOnly);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    // read data
    int nRows = poRAT->GetRowCount();
    if( nRows != nNumRows )
        errexit("wrong number of rows");

    // only check the first 10
    nRows = 10;

    int *panCheckData = (int*)CPLCalloc(sizeof(int), nRows);
    if( poRAT->ValuesIO(GF_Read, 0, 0, nRows, panCheckData ) != CE_None )
        errexit("Failed to read int column");
    if( !CheckArraysInt(panCheckData, anIntData, nRows) )
        errexit("Int colunm does not match");

    double *padfCheckData = (double*)CPLCalloc(sizeof(double), nRows);
    if( poRAT->ValuesIO(GF_Read, 1, 0, nRows, padfCheckData ) != CE_None )
        errexit("Failed to read double column");
    if( !CheckArraysDouble(padfCheckData, adfDoubleData, nRows) )
        errexit("Double colunm does not match");

    char **papszCheckData = (char**)CPLCalloc(sizeof(char*), nRows);
    if( poRAT->ValuesIO(GF_Read, 2, 0, nRows, papszCheckData ) != CE_None )
        errexit("Failed to read string column");
    if( !CheckArraysString(papszCheckData, apszStringData, nRows) )
        errexit("String colunm does not match");
    for( int i = 0; i < nRows; i++ )
        CPLFree(papszCheckData[i]);


    if( poRAT->ValuesIO(GF_Read, 3, 0, nRows, padfCheckData) != CE_None)
        errexit("Failed to read int as double column");
    if( !CheckArraysDoubleAsInt(padfCheckData, adfDoubleData, nRows) != CE_None)
        errexit("int as double column does not match");

    if( poRAT->ValuesIO(GF_Read, 4, 0, nRows, papszCheckData) != CE_None)
        errexit("Failed to read int as string column");
    if( !CheckArraysString(papszCheckData, apszStringDataInts, nRows))
        errexit("int as string column does not match");
    for( int i = 0; i < nRows; i++ )
        CPLFree(papszCheckData[i]);

    if( poRAT->ValuesIO(GF_Read, 5, 0, nRows, panCheckData) != CE_None)
        errexit("Failed to read double as int column");
    if( !CheckArraysInt(panCheckData, anIntData, nRows))
        errexit("double as int column does not match");

    if( poRAT->ValuesIO(GF_Read, 6, 0, nRows, papszCheckData) != CE_None)
        errexit("Failed to read double as string column");
    if( !CheckArraysStringAsDouble(papszCheckData, apszStringDataDoubles, nRows))
        errexit("double as string column does not match");

    CPLFree(panCheckData);
    CPLFree(padfCheckData);
    CPLFree(papszCheckData);

    fprintf( stderr, "Values Matched\n" );

    GDALClose( (GDALDatasetH) poDS );
}

void CheckSetGetValues(const char *pszName)
{
    // check the 'legacy' getvalue/setvalue calls
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    // write data
    int nRows = poRAT->GetRowCount();

    for( int i = 0; i < nRows; i++)
    {
        // write some data slightly differrent
        poRAT->SetValue(i, 0, anIntData[i] + 1);
        poRAT->SetValue(i, 1, adfDoubleData[i] + 1);
        char *pszBuffer = CPLStrdup(apszStringData[i]);
        if( strlen(pszBuffer) > 0 )
            pszBuffer[0] = 'z';
        poRAT->SetValue(i, 2, pszBuffer);
        CPLFree(pszBuffer);
    }

    // read data and check
    for( int i = 0; i < nRows; i++)
    {
        if( poRAT->GetValueAsInt(i, 0) != (anIntData[i] + 1))
            errexit("GetValueAsInt not reading correctly");
        if( poRAT->GetValueAsDouble(i, 1) != ( adfDoubleData[i] + 1))
            errexit("GetValueAsDouble not reading correctly");

        char *pszBuffer = CPLStrdup(apszStringData[i]);
        if( strlen(pszBuffer) > 0 )
            pszBuffer[0] = 'z';
        if( !EQUAL(pszBuffer, poRAT->GetValueAsString(i, 2)))
            errexit("GetValueAsString not reading correctly");
        CPLFree(pszBuffer);
    }

    // no need to check different types as ValuesIO is checked for this above
    // and these calls map to ValuesIO

    // write back old
    for( int i = 0; i < nRows; i++)
    {
        poRAT->SetValue(i, 0, anIntData[i]);
        poRAT->SetValue(i, 1, adfDoubleData[i]);
        poRAT->SetValue(i, 2, apszStringData[i]);
    }

    fprintf( stderr, "Get/SetValue OK\n");

    GDALClose( (GDALDatasetH) poDS );
}

void ExtendAndWrite(const char *pszName)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    int nOldRows = poRAT->GetRowCount();

    // extend
    int nNewLen = sizeof(anIntData) / sizeof(int);
    int nRows = nOldRows + nNewLen;
    poRAT->SetRowCount(nRows);

    // write new
    if( poRAT->ValuesIO(GF_Write, 0, nOldRows, nNewLen, anIntData) != CE_None )
        errexit("Failed to write int column");
    if( poRAT->ValuesIO(GF_Write, 1, nOldRows, nNewLen, adfDoubleData) != CE_None )
        errexit("Failed to write double column");
    if( poRAT->ValuesIO(GF_Write, 2, nOldRows, nNewLen, apszStringData) != CE_None )
        errexit("Failed to write string column");

    GDALClose( (GDALDatasetH) poDS );
}

void CheckExtension(const char *pszName)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();
    int nRows = 10;

    int *panCheckData = (int*)CPLCalloc(sizeof(int), nRows);
    if( poRAT->ValuesIO(GF_Read, 0, 10, nRows, panCheckData ) != CE_None )
        errexit("Failed to read int column");
    if( !CheckArraysInt(panCheckData, anIntData, nRows) )
        errexit("Int colunm does not match");

    double *padfCheckData = (double*)CPLCalloc(sizeof(double), nRows);
    if( poRAT->ValuesIO(GF_Read, 1, 10, nRows, padfCheckData ) != CE_None )
        errexit("Failed to read double column");
    if( !CheckArraysDouble(padfCheckData, adfDoubleData, nRows) )
        errexit("Double colunm does not match");

    char **papszCheckData = (char**)CPLCalloc(sizeof(char*), nRows);
    if( poRAT->ValuesIO(GF_Read, 2, 10, nRows, papszCheckData ) != CE_None )
        errexit("Failed to read string column");
    if( !CheckArraysString(papszCheckData, apszStringData, nRows) )
        errexit("String colunm does not match");
    for( int i = 0; i < nRows; i++ )
        CPLFree(papszCheckData[i]);

    fprintf(stderr, "extension values ok\n");
    GDALClose( (GDALDatasetH) poDS );
}

// this will force the string column to be re-written to accomodate
// a longer string size
void WriteLongStrings(const char *pszName)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    if( poRAT->ValuesIO(GF_Write, 2, 10, 10, apszLongStringData) != CE_None )
        errexit("Failed to write string column");

    fprintf(stderr, "wrote long strings ok\n");
    GDALClose( (GDALDatasetH) poDS );
}

void CheckLongStrings(const char *pszName)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    char **papszCheckData = (char**)CPLCalloc(sizeof(char*), 10);
    if( poRAT->ValuesIO(GF_Read, 2, 10, 10, papszCheckData ) != CE_None )
        errexit("Failed to read string column");
    if( !CheckArraysString(papszCheckData, apszLongStringData, 10) )
        errexit("String colunm does not match");
    for( int i = 0; i < 10; i++ )
        CPLFree(papszCheckData[i]);

    fprintf(stderr, "checked long strings ok\n");
    GDALClose( (GDALDatasetH) poDS );
}

void SetLinearBinning(const char *pszName)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    if( poRAT->SetLinearBinning(0, 1) != CE_None)
        errexit("Error in SetLinearBinning");

    fprintf(stderr, "set linear binning ok ok\n");
    GDALClose( (GDALDatasetH) poDS );
}

void CheckLinearBinning(const char *pszName)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    double dfMin, dfSize;
    if( !poRAT->GetLinearBinning(&dfMin, &dfSize) )
        errexit("GetLinearBinning failed");

    if( ( dfMin != 0 ) || (dfSize != 1))
        errexit("GetLinearBinning values wrong");

    if( poRAT->GetRowOfValue(3) != 3)
        errexit("GetRowOfValue value wrong");

    fprintf(stderr, "linear binning ok ok\n");
    GDALClose( (GDALDatasetH) poDS );
}

void CheckClone(const char *pszName)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    GDALDefaultRasterAttributeTable *poCloned = poRAT->Clone();

    if(poCloned->GetValueAsInt(0, 0) != 197)
        errexit("Cloned info wrong int");
    if(poCloned->GetValueAsDouble(5, 1) != 0.92)
        errexit("Cloned info wrong double");
    if(!EQUAL(poCloned->GetValueAsString(1, 2), "wess"))
        errexit("Cloned info wrong string");

    delete poCloned;

    fprintf(stderr, "clone ok\n");
    GDALClose( (GDALDatasetH) poDS );
}

void CheckSerialize(const char *pszName)
{
    GDALDataset *poDS = (GDALDataset*)GDALOpen(pszName, GA_Update);

    GDALRasterBand *poBand = poDS->GetRasterBand(1);

    // Get the RAT
    GDALRasterAttributeTable *poRAT = poBand->GetDefaultRAT();

    // serialize
    CPLXMLNode *pNode = poRAT->Serialize();

    // create a new one and init from xml
    GDALDefaultRasterAttributeTable oCloned;
    oCloned.XMLInit(pNode, NULL);

    if(oCloned.GetValueAsInt(0, 0) != 197)
        errexit("Cloned info wrong int");
    if(oCloned.GetValueAsDouble(5, 1) != 0.92)
        errexit("Cloned info wrong double");
    if(!EQUAL(oCloned.GetValueAsString(1, 2), "wess"))
        errexit("Cloned info wrong string");

    fprintf(stderr, "serialize ok\n");
    GDALClose( (GDALDatasetH) poDS );
}


int main()
{
    GDALAllRegister();

    VSIUnlink("test.img");
    
    // basic tests
    CreateAndWriteRAT("test.img");

    ReadAndCheckValues("test.img", 10);

    // the older interface
    CheckSetGetValues("test.img");

    // make sure original data not changed
    ReadAndCheckValues("test.img", 10);

    // make it longer - data will be re-written
    ExtendAndWrite("test.img");
    // make sure old data not changed
    ReadAndCheckValues("test.img", 20);

    // new data at the end ok?
    CheckExtension("test.img");

    // write some longer strings - string column will 
    // have to be re-written
    WriteLongStrings("test.img");

    // make sure old data not changed
    ReadAndCheckValues("test.img", 20);

    // check new data ok
    CheckLongStrings("test.img");

    // linear binning
    SetLinearBinning("test.img");
    CheckLinearBinning("test.img");

    // clone
    CheckClone("test.img");

    // serialize
    CheckSerialize("test.img");

    fprintf(stderr, "All ok\n" );

    return 0;
}

